/* See LICENSE file for copyright and license details. */
#include <err.h>
#include <errno.h>
#include <dirent.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/stat.h> 
#include <time.h>
#include <unistd.h> /* pledge */

#include "config.h"

#define PLEDGEORDIE(prom) if (pledge(prom, NULL) == -1) { err(1, "pledge"); } 
#define UNVEILORDIE(path,perms) if (unveil(path, perms) == -1) { err(1, "unveil"); } 

char *       datetime(const char *fmt, time_t clock);
void *       ecalloc(size_t nmemb, size_t size);
int          efclose(FILE *stream);
FILE *       efopen(const char *path, const char *mode);
int          efprintf(FILE *stream, const char *format, ...);
char *       estrdup(const char *s);
void *       ereallocarray(void *ptr, size_t nmemb, size_t size);
size_t       esnprintf(char *str, size_t size, const char *fmt, ...);
char *       get_article(const char *fp);
char *       get_cmd_output(const char *cmd);
char *       get_first_line(const char *f);
char *       get_file_txt(const char *f);
struct stat  get_lastmod(const char *fp);
char *       get_stream_txt(FILE *fp);
char *       get_title(const char *f);
int          isfile(const char *path);
void         mkatom(int idx, const char *dir);
void         mkindex(const char *dir, int nbpage);
void         mkpage(int n, const char *dir);
void         mksitemap(const char *dir, int nbpage);
int          num_files(const char *path, const char *ext);


void *
ecalloc(size_t nmemb, size_t size)
{
	void *p;


	if ((p = calloc(nmemb, size)) == NULL)
		err(1, "calloc");
	return p;
}

void *
ereallocarray(void *ptr, size_t nmemb, size_t size)
{
	if ((ptr = reallocarray(ptr, nmemb, size)) == NULL)
		err(1, "reallocarray");

	return ptr;
}

int
efclose(FILE *stream)
{
	int ret = 0;


	ret = fclose(stream);
	if (ferror(stream)) {
		err(1,"closefile");
	}
	return ret;
}

FILE *
efopen(const char *path, const char *mode)
{
	FILE *f = NULL;


	if ((f = fopen(path, mode)) == NULL)
		err(1, "%s: %s", "Fail to open", path);

	return f;
}

int
efprintf(FILE *stream, const char *format, ...)
{
	va_list ap;
	int ret;


	va_start(ap, format);
	ret = vfprintf(stream, format, ap);
	if (ret < 0)
		err(1,"vfprintf:");
	va_end(ap);

	return ret;
}

char *
estrdup(const char *s)
{
	char *p = strdup(s);


	if (p == NULL)
		err(1, "strdup");

	return p;
}

size_t
esnprintf(char *str, size_t size, const char *fmt, ...)
{
    va_list ap;
    size_t ret = 0;


    va_start(ap, fmt);
    ret = vsnprintf(str, size, fmt, ap);
    if (ret < 0) {
        err(1,"vsnprintf:");
    } else if (ret >= size) {
        err(1,"vsnprintf: Output truncated");
	}

    va_end(ap);

    return ret;
}

char *
get_first_line(const char *f)
{
	char line[BUFSIZ] = {0};
	char c;
	int idx = 0;
	FILE *fp;


	fp = efopen(f, "r");

	while ((c = fgetc(fp)) != EOF) {
		if (c == '\n')
			break;
		line[idx] = c;
		idx++;
	}
	line[idx+1] = '\0';
	efclose(fp);

	return estrdup(line);
}

char *
get_title(const char *f)
{
	char *title = NULL;

	title = get_first_line(f);

	/* trim first '#' and spaces */
	while ((title[0] == ' ') || (title[0] == '#')) {
		memmove(title, title+1, strlen(title));
	}

	return title;

}

/* return text stored in f */
char *
get_file_txt(const char *f)
{
	char *str = NULL;
	FILE *fp;
	fp = efopen(f, "r");
	str = get_stream_txt(fp);
	efclose(fp);

	return str;
}

/* return text in fp
 * the fp must be closed properly after
 */
char *
get_stream_txt(FILE *fp)
{
	char *buf = NULL;
	size_t s;
	size_t len = 0;
	size_t bsize = 2 * BUFSIZ;


	buf = ecalloc(bsize, sizeof(char));
	while((s = fread(buf + len, 1, BUFSIZ, fp))) {
		len += s;
		if(BUFSIZ + len + 1 > bsize) {
			bsize += BUFSIZ;
			buf = ereallocarray(buf, bsize, sizeof(char));
		}
	}
	buf[len] = '\0';

	return(buf);
}

int
num_files(const char *path, const char *ext)
{
	struct dirent *dp;
	DIR *fd;
	int num = 0;
	char *dot;


	if (!(fd = opendir(path))) {
		err(1,"opendir '%s':", path);
	}

	while ((dp = readdir(fd))) {
		if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, "..")) {
			continue; /* skip self and parent */
		}
		dot = strrchr(dp->d_name, '.');

		if (dot != NULL) {
			if (strcmp(dot, ext) == 0) {
				num++;
			}
		}
	}
	closedir(fd);

	return num;
}

int
isfile(const char *path)
{
	if (access(path, R_OK ) != -1) {
		return 1;
	} else {
		return 0;
	}
}

char *
get_cmd_output(const char *cmd)
{
	FILE *fp = NULL;
    char *res = NULL;

    if ((fp = popen(cmd, "r")) == NULL)
        err(1, "%s: %s", "Error opening pipe for cmd", cmd);

    res = get_stream_txt(fp);
    if (pclose(fp) != 0) {
        err(1, "%s %s", cmd, "not found or exited with error status");
	}
	
	return res;
}

char *
get_article(const char *fp)
{
	char *txt = NULL;
	char cmd[BUFSIZ] = {'\0'};


	if (markup == NULL) {
		txt = get_file_txt(fp);
	} else {
		esnprintf(cmd, sizeof(cmd), "%s %s", markup, fp);
		txt = get_cmd_output(cmd);
	}

	return txt;
}

void
mkpage(int n, const char *dir)
{
	char fp[PATH_MAX] = {'\0'};
	char hfp[PATH_MAX] = {'\0'};
	char prevp[PATH_MAX] = {'\0'};
	char nextp[PATH_MAX] = {'\0'};
	char arrowlinks[120] = {'\0'};
	char *txt = NULL;
	int nblinks = 0;
	FILE *f;


	/* file names */
	esnprintf(fp, sizeof(fp), "%s/%d.txt", dir, n);
	esnprintf(hfp, sizeof(hfp), "%s/%d.html", dir, n);

	/* link to previous and next pages */
	esnprintf(prevp, sizeof(prevp), "%s/%d.txt", dir, n-1);
	esnprintf(nextp, sizeof(nextp), "%s/%d.txt", dir, n+1);
	nblinks = isfile(prevp) + isfile(nextp);
	switch (nblinks) {
	case 0: 
		break;
	case 1: /* previous except for #1 */
		if (n == 1) {
			esnprintf(arrowlinks, sizeof(arrowlinks), 
				"<a href=\"%d.html\" class=\"arrow next\">&gt;</a>", n+1);
		} else {
			esnprintf(arrowlinks, sizeof(arrowlinks), 
				"<a href=\"%d.html\" class=\"arrow prev\">&lt;</a>", n-1);
		}
		break;
	case 2: /* previous and next*/
		esnprintf(arrowlinks, sizeof(arrowlinks), 
			"<a href=\"%d.html\" class=\"arrow prev\">&lt;</a>\n\
<a href=\"%d.html\" class=\"arrow next\">&gt;</a>", 
			n-1, n+1);
		break;
	default:
		break;
	}

	txt = get_article(fp);
	f = efopen(hfp, "w");
	/* write page with head and foot */
	efprintf(f, headhtml, n, n, n);
	efprintf(f, "%s\n%s\n%s", txt, arrowlinks, footerhtml);
	efclose(f);
	free(txt);
}

void
mkatom(int idx, const char *dir)
{

	char fp[PATH_MAX] = {'\0'};
	char *updated = NULL, *title = NULL, *txt = NULL;
	FILE *fatom;

	/* do atom */
	/* path to atom file and open it*/
	esnprintf(fp, sizeof(fp), "%s/%s", dir, atomfp);
	fatom = efopen(fp, "w");

	/* insert updated time */
	updated = datetime("%Y-%m-%dT%TZ", 0);
	efprintf(fatom, atom, url, url, url, url, updated);
	free(updated);

	/* check every txt file */
	/*
	int stop = 0;
	if (idx > maxentry) {
		stop = idx - maxentry;
	} else {
		stop = 0;
	}
	*/
	int stop = (idx > maxentry) ? idx - maxentry : 0;

	for (int i = idx; i > stop; i--) {
		esnprintf(fp, sizeof(fp), "%s/%d.txt", dir, i);
		updated = datetime("%Y-%m-%dT%TZ", get_lastmod(fp).st_mtime);

		title = get_title(fp);
		txt = get_article(fp);
		efprintf(fatom, atomentry, url, i, title, url, i, updated, txt);

		free(title);
		free(txt);
		free(updated);
	}

	/* end and close atom */
	efprintf(fatom, "%s", "</feed>");
	efclose(fatom);
}

struct stat
get_lastmod(const char *fp)
{
	/*memset( &s, 0, sizeof(struct stat) );*/
	struct stat lastmod = {.st_mtime = 0};

	if (stat(fp, &lastmod) != 0) {
		warn("Can't find mod time of:%s", fp); 
	}

	return lastmod;
}

char *
datetime(const char *fmt, time_t clock)
{
	char date[255] = {'\0'}; /* this is too big */

	if (clock == 0) {
		clock = time(NULL);
	}

	strftime(date, sizeof(date), fmt, localtime(&clock));

	return estrdup(date);
}

void
mkindex(const char *dir, int nbpage)
{
	FILE *indexfp;
	char fp[PATH_MAX]; 
	char a[PATH_MAX]; 
	char *title = NULL;

	esnprintf(fp, sizeof(fp), "%s/index.html", dir);
	indexfp = efopen(fp, "w");

	efprintf(indexfp, indexheadhtml);

	efprintf(indexfp, "<ul id=\"proseartlist\">");
	/* loop on all articles pages */
	for (int i = nbpage; i > 0; i--) {
		esnprintf(a, sizeof(a), "%s/%d.txt", dir, i);
		title = get_title(a);
		efprintf(indexfp, "<li><a href=\"%d.html\">%d : %s</a>",
							i, i, title); 
		free(title);
	}

	efprintf(indexfp, "</ul>");
	efprintf(indexfp, indexfooterhtml);

	efclose(indexfp);
}

void
mksitemap(const char *dir, int nbpage)
{
	FILE *sitemapfp;
	char fp[PATH_MAX]; 

	esnprintf(fp, sizeof(fp), "%s/sitemap.xml", dir);
	sitemapfp = efopen(fp, "w");

	efprintf(sitemapfp, "%s\n%s", "<?xml version=\"1.0\" encoding=\"UTF-8\"?>",
		"<urlset xmlns=\"http://www.sitemaps.org/schemas/sitemap/0.9\">\n");

	/* loop on all articles pages */
	for (int i = nbpage; i > 0; i--) {
		efprintf(sitemapfp, "<url>\n  <loc>%s/%d.html</loc>\n</url>\n", url, i); 
	}

	efprintf(sitemapfp,"%s", "</urlset>");

	efclose(sitemapfp);
}

int 
main(int argc, char *argv[])
{
	if (argc != 2) {
		printf("usage:\n");
		printf("\t argv[0] <output directory>\n");
		return 1;
	}

#ifdef __OpenBSD__
	if (markup == NULL) {
		UNVEILORDIE(argv[1], "rwc");
		PLEDGEORDIE("stdio rpath cpath wpath");
	} else {
		UNVEILORDIE(argv[1], "rwc");
		UNVEILORDIE(markup, "x");
		UNVEILORDIE("/bin/sh", "x");
		PLEDGEORDIE("stdio rpath cpath wpath exec proc");
	}
#endif

	char *line = NULL, *mail = NULL, *txt = NULL;
	
	char fp[PATH_MAX] = {'\0'};
	FILE *f;
	int idx = 0, n = 0;

	/* store message in memory. If memory is not enough, you're 
	 * probably insane to write this much */
	mail = get_stream_txt(stdin);

	/* remove header */
	if ((txt = strstr(mail, "\n\n")) == NULL) {
		err(1, "%s", "Badly formatted mail, exit");
	}
	txt = txt + 2; /* skip "\n\n"; */

	/* message start after \n\n */
	/* get first line and check if it is the password */
	line = strsep(&txt, "\n");
	if (line == NULL) {
		err(1, "%s", "no password found, exit");
	}
	if (strcmp(line, pw) != 0) {
		err(1, "%s", "wrong password, exit");
	}

	/* we can read the rest of the body */
	/* don't care of firsts \n */
	while (strcspn(txt, "\n") == 0) {
		txt = txt +1;
	}

	/* get number of articles */
	n = num_files(argv[1], ext);

	/* store in n+1.txt */
	idx = n + 1;
	esnprintf(fp, sizeof(fp), "%s/%d.txt", argv[1], idx);
	f = efopen(fp, "w");
	efprintf(f, "%s", txt);
	efclose(f);

	free(mail);

	/* now create html pages */
	mkpage(idx, argv[1]);
	/* previous page to add links */
	if (idx >= 2) {
		mkpage(idx-1, argv[1]);
	}

	mkindex(argv[1], idx);
	mkatom(idx, argv[1]);
	mksitemap(argv[1], idx);

	return 0;
}
